
#include "shading.h"
#include "color.h"
#include "reflection.h"
#include "texture.h"
#include "geometry.h"
#include "butterfly.h"


void WingSurface::GenerateUpper(){
  RGBATex *tex, *base;
  Transform scale, move, rot;

  printf( "Generating Upper Wing Surface\n" );

  Spectrum defaultColor = bgeo->wingBaseColor; //Spectrum( .5, .5, .5 );
  wingTex = new TextureMap( 500, 500, defaultColor );

  base = new RGBATex( "wings/plainupperwing.tif", "wings/plainupperwing_alpha.tif" );
  scale = Scale( 1.5, 1.5, 1.5 );
  move = Translate( Vector( .5, .5, 0 ) );
  rot = Rotate( 0, Vector( 0, 0, 1) );
  TransformTex( base, move, scale, rot );




  switch( bgeo->textureMode ){
  case 0:
    break;
    
  case 1:
    base = new RGBATex( "wings/plainupperwing.tif", "wings/plainupperwing_alpha.tif" );
    break;
  case 2:
    base = new RGBATex( "scales/scales1.tif", "noalpha.tif" );
    break;
  case 3:
    base = new RGBATex( "scales/scales2.tif", "noalpha.tif" );
    break;
  case 4:
    base = new RGBATex( "scales/scales3.tif", "noalpha.tif" );
    break;

  }

  base->repeat = true;
  scale = Scale( .5, .5, .5 );
  move = Translate( Vector( .5, .5, 0 ) );
  rot = Rotate( 0, Vector( 0, 0, 1) );
  TransformTex( base, move, scale, rot );

  if( bgeo->textureMode > 0 )
    Composite( wingTex, base, OVERLAY );


  //Composite( wingTex, base, OVERLAY );
  tex = new RGBATex( "spots/bluespot.tif", "spots/bluespot_alpha.tif" );
  //Composite( wingTex, tex, OVERLAY  );
  
  scale = Scale( .4, .4, 1 );
  move = Translate( Vector( 0.5, 0.5, 0 ) );
  rot = Rotate( 0, Vector( 0, 0, 1) );
  
  TransformTex( tex, move, scale, rot );
  //Composite( wingTex, tex, OVERLAY  );

 
  //GenerateRow( tex, 0.3, 0.4 );
  GenerateSpots();

 
  DrawAllVeins();
}


void WingSurface::GenerateSpots(){
  RGBATex *tex;

   
  //Generate Rows
  printf( "Generating Spots...\n" );
  float spacingFactor = 0.8;
  int numRows = RandomRange( 1, 5 );
  float spotUstep = 1.0 / (float)(numRows+1);
  float spotCurUstep = 0.0;
  for( int i = 0; i < numRows; i++ ){
    int spotNum = rand() % texMasks.size();
    float spotSize = 0.5; //RandomRange( .01, 0.5 );
    //tex = texSpots[spotNum];
    int maskNum = rand() % texMasks.size();
    tex = texMasks[maskNum];
    printf( "Generating row tex: %d u: %f size: %f\n", spotNum,
    spotCurUstep, spotSize );
    tex->repeat = false;
    GenerateRow( tex, spotCurUstep, spotSize, spotUstep*spacingFactor );
    spotCurUstep += spotUstep;
  }

  // Big Spots
  int numBigSpots = 2; //RandomRange( 0, 2 );
  for( int j = 0; j < numBigSpots; j++ ){
    float u = RandomRange( 0, 1 );
    float v = RandomRange( 0, 1 );
    int spotNum = rand() % texSpots.size();
    float spotSize = RandomRange( .01, 0.7 );
    Transform scale = Scale( spotSize, spotSize, spotSize );
    Transform move = Translate( u, v, 0 );
    Transform rot = Rotate( 0, Vector( 0, 0, 1) );
    tex = texSpots[spotNum];
    tex->repeat = false;
    TransformTex( tex, move, scale, rot );
    Composite( wingTex, tex, OVERLAY );
    tex = texSpots[spotNum];
  }

  
}

void WingSurface::GenerateRow( RGBATex *tex, float u, float width, float height ){
  
  int skip = RandomRange( 0.0, 2.0 );
  for( int i = 0; i < wgeo->outline->numOuterEdgeVeins-1; i++ ){
    //if( skip && isOdd(i) )
    //continue;
    GenerateSectorSpot( tex, u, u + height, 0.5 - width, 0.5 + width, i );
  }
}


void WingSurface::GenerateLower(){
 RGBATex *tex, *base;
  Transform scale, move, rot;

  printf( "Generating Lower Wing Surface\n" );

  Spectrum defaultColor = bgeo->wingBaseColor; //Spectrum( .5, .5, .5 );
  wingTex = new TextureMap( 500, 500, defaultColor );

  base = new RGBATex( "wings/plainlowerwing.tif", "wings/plainlowerwing_alpha.tif" );
  base = new RGBATex( "scales/scales2.tif", "noalpha.tif" );
 

  switch( bgeo->textureMode ){
  case 0:
    break;

  case 1:
    base = new RGBATex( "wings/plainlowerwing.tif", "wings/plainlowerwing_alpha.tif" );
    break;
  case 2:
    base = new RGBATex( "scales/scales1.tif", "noalpha.tif" );
    break;
  case 3:
    base = new RGBATex( "scales/scales2.tif", "noalpha.tif" );
    break;
  case 4:
    base = new RGBATex( "scales/scales3.tif", "noalpha.tif" );
    break;

  }

  base->repeat = true;
  scale = Scale( .5, .5, .5 );
  move = Translate( Vector( .5, .5, 0 ) );
  rot = Rotate( 0, Vector( 0, 0, 1) );
  TransformTex( base, move, scale, rot );

  if( bgeo->textureMode > 0 )
    Composite( wingTex, base, OVERLAY );


  tex = new RGBATex( "spots/bluespot.tif", "spots/bluespot_alpha.tif" );
  Composite( wingTex, tex, OVERLAY  );
  
  scale = Scale( .4, .4, 1 );
  move = Translate( Vector( 0.5, 0.5, 0 ) );
  rot = Rotate( 0, Vector( 0, 0, 1) );
  
  TransformTex( tex, move, scale, rot );
  //Composite( wingTex, tex, OVERLAY  );

  //GenerateSectorSpot( 0.5, 0.85, 0.0, 1.0, 1 ); 
  //GenerateSectorSpot( 0.5, .85, 0.0, 1.0, 2 );
  //GenerateSectorSpot( 0.0, 1.0, 0.0, 1.0, 3 );

  

  //PlaceOnSector( wingTex, tex, 0.5f, 0.5f, 2, scale, rot );
  //PlaceOnSector( wingTex, tex, 0.5f, 0.5f, 4, scale, rot );

  GenerateSpots();

  DrawAllVeins();

}

void WingSurface::Mirror( WingSurface *mirror ){
  //Spectrum color;
  //Spectrum defaultColor = Spectrum( .5, .5, .5 );
  //int width = mirror->wingTex->GetWidth();
  //int height = mirror->wingTex->GetHeight();

  wingTex = mirror->wingTex;

  // old way, now done in Shade
  /*wingTex = new TextureMap( width, height, defaultColor );

  for( int x = 0; x < width; x++ ){
    for( int y = 0; y < height; y++ ){
      color = mirror->wingTex->LookupScreen( height-1-x, y );
      wingTex->Set( x, y, color );

    }
    }*/

}


void WingSurface::Generate( char* ident, WingSurface *mirror /*=NULL*/ ){

  identity = ident;

  if( ! bgeo ){
    printf( "WingSurface called with out butterfly primitive\n" );
    return;
  }

  if( identity[1] == 'l' && mirror == NULL ){
    printf( "Left wing called without right wing to mirror\n" );
    exit(0);
  }
 
  if( identity[1] == 'r' ){
    if( identity[0] == 'l' ){
      wgeo =  bgeo->llWing;
      GenerateLower();
    }
    if( identity[0] == 'u' ){
      wgeo =  bgeo->ulWing;
      GenerateUpper();
    }
  }
  else if( identity[1] == 'l' ){
    Mirror( mirror );
  }

}

void WingSurface::PlaceOnSector( TextureMap *basetex, RGBATex *src, 
				 float sectorU, float sectorV,
				 int secNum, Transform &scale,
				 Transform &rot ){
     
  //float minu = wgeo->outline->outerEdgeVeins[secNum]->minu;

  //if( sectorU < minu )
  //  return;

  Point *pos = wgeo->outline->RStoWS( secNum, sectorU, sectorV );

  if( ! pos ){
    printf( "NULL on RStoWS\n");
    return;
  }

  //basetex->ToRasterCoords( pos.x, pos.y, &newposx, &newposy );
  Transform move = Translate(pos->x, pos->y, pos->z);
  TransformTex( src, move, scale, rot );
  Composite( basetex, src, OVERLAY );

  printf( "PlacceOnSector: %f %f\n", pos->x, pos->y );


}

void WingSurface::GenerateSectorSpot( RGBATex *src, float uStart, float uEnd,
				      float vStart, float vEnd,
				      int secNum ){
  float srcU, srcV;
  float step = .002;
  Point *wingUV;
  double srcUstep = 1/(uEnd - uStart)*step;
  double srcVstep = 1/(vEnd - vStart)*step;
  float spotU = 0, spotV = 0;

  src->trans = Identity();

  float minu = wgeo->outline->outerEdgeVeins[secNum+1]->minu;

   
  for( srcU = uStart; srcU <= uEnd; srcU+= step ){
    spotV = 0;
    for( srcV = vStart; srcV <= vEnd; srcV+= step ){

      
      wingUV = wgeo->outline->RStoWS( secNum, srcU, srcV );
      if( ! wingUV ){
	printf( "Ustart too small on secNum %d   umin:  %f ustart: %f\n", secNum, minu, uStart );
	return;
      }

      
      //CompositePixel( wingTex, src, Point( srcU, srcV, 0), wingUV, OVERLAY );
      CompositePixel( wingTex, src, Point( spotU, spotV, 0 ),
		      Point(wingUV->x,wingUV->y,0), OVERLAY );
      spotV += srcVstep;
    }
    spotU += srcUstep;
  }
  printf( "spot %d %f %f\n", secNum, spotU, spotV );
  
}


void WingSurface::DrawAllVeins(){
  
  printf("DrawingVeins\n");
  int numVeins;
 
  numVeins = wgeo->outline->numOuterEdgeVeins;
  
 
  for( int i = 0; i < numVeins; i++ )
    DrawVein( i );
    
}


void WingSurface::DrawVein( int veinNum ){
  printf( "Drawing vein %d\n", veinNum );

  Spectrum color = Spectrum( 0, 0, 0 );
  float minu;
  float step = .0002;
  Vein** veins;

  /*if( identity[0] == 'u' )
    veins = bgeo->ulWing->outline->outerEdgeVeins;
  else
  veins = bgeo->llWing->outline->outerEdgeVeins;*/

  veins = wgeo->outline->outerEdgeVeins;
  minu = wgeo->outline->outerEdgeVeins[veinNum]->minu;

  for( float vU = minu; vU <= 1.0; vU+= step ){
    Point vP = veins[veinNum]->GetVeinPoint(vU);
    //lowerwing->SetRaster( vP.x, vP.y, color );
    DrawSquare( vP, .005, .005, color );
  }
}

// all in raster units [0 1]
void WingSurface::DrawSquare( Point center, float width, float height,
			      Spectrum &color){
  float hwidth = width/2.0;
  float hheight = height/2.0;
  float step = .002;
  Spectrum baseColor;
  Spectrum finalColor;
  
  for( float x = -hwidth; x < hwidth; x+= step /*wingTex->xStep()*/ ){
    for( float y = -hheight; y < hheight; y+= step /*wingTex->yStep()*/ ){
      if( center.x + x > 1.0 ||center.y + y > 1.0 ) continue;
      if( center.x + x < 0.0 ||center.y + y < 0.0 ) continue;
      baseColor = wingTex->Lookup( center.x + x, center.y + y, false );
      finalColor = (color + baseColor)/2;
      wingTex->SetRaster( center.x + x, center.y + y, finalColor );
    }
  }
}


void WingSurface::TransformTex( RGBATex *tex, Transform &move, Transform
&scale, Transform &rot ){
    
    Transform center = Translate( -.5, -.5, 0 );
    //Transform centerInv = Translate( 0.5, 0.5, 0 );

    tex->trans = move * rot * scale * center;
  
}


void WingSurface::CheckBounds( float u, float v ){
  if( u > 1.0f || v > 1.0f || u < 0.0 | v < 0.0 ){
    printf( "\n\nWarning, wrapping or clamping may occur. u: %f v: %f\n",
	   u, v );
    assert( false );
  }  
}

void WingSurface::CompositePixel( TextureMap* baseTex, RGBATex* src, Point srcUV, Point dstUV,
compositeMode mode /*=OVERLAY*/ ){
  int x, y;

  baseTex->ToScreenCoords( dstUV.x, dstUV.y, &x, &y );

  CompositePixel( baseTex, src, srcUV, x, y, mode );

}

void WingSurface::CompositePixel( TextureMap* baseTex, RGBATex* src, Point srcUV, int dstX, int dstY,
				  compositeMode mode /*=OVERLAY*/ ){

  Point dstUV;
  float srcAlpha;
  TextureMap *srcTmap, *srcAmap;
  Spectrum srcColor, dstColor, finalColor;
 
  baseTex->ToRasterCoords( dstX, dstY, &dstUV.x, &dstUV.y );

  srcTmap = ShadeContext::GetTexture( src->texture ); 
  srcAmap = ShadeContext::GetTexture( src->alpha );
  
  srcColor = srcTmap->Lookup( srcUV.x, srcUV.y, src->repeat );
  srcAlpha = srcAmap->Lookup( srcUV.x, srcUV.y, src->repeat )[0];
  dstColor = baseTex->Lookup( dstUV.x, dstUV.y, false );
 
  //CheckBounds( srcUV.x, srcUV.y );
  //CheckBounds( dstUV.x, dstUV.y );
  
  switch( mode ){
  case OVERLAY:
    finalColor = (1-srcAlpha) * dstColor + srcAlpha * srcColor;
    break;
  case BLEND:
    finalColor = 0.5 * dstColor + 0.5 * srcAlpha * srcColor;
    break;
  }
  
  baseTex->Set( dstX, dstY, finalColor );
  
}


void WingSurface::Composite(TextureMap* baseTex, RGBATex* src, compositeMode mode /*=OVERLAY*/ ){
  int x, y;
  float rX, rY;
  TextureMap *srcTmap, *srcAmap;
 
  srcTmap = ShadeContext::GetTexture( src->texture ); 
  srcAmap = ShadeContext::GetTexture( src->alpha );
  
  for( x = 0; x < baseTex->GetWidth(); x++ ){
    for( y = 0; y < baseTex->GetHeight(); y++ ){
      baseTex->ToRasterCoords( x, y, &rX, &rY );
      

      Point srcST( rX, rY, 0 );
     
      Transform srcSTinv = src->trans.GetInverse();
      srcST = srcSTinv( srcST );
     
      CompositePixel( baseTex, src, srcST, x, y, mode );

      /*printf("Before: %d, %d, %f, %f\n", x, y, rX, rY );
      int testx, testy;
      baseTex->ToScreenCoords( rX, rY, &testx, &testy );
      printf("After: %d, %d\n", testx, testy );
      assert( testx == x && testy == y );*/

      //CompositePixel( baseTex, src, srcST, Point( rX, rY), mode );
    }
  }
}


BRDF *WingSurface::Shade(const ShadeContext & sc) const
{
  // TextureMap *tmap;
  Float s, t;

  
  // wing
  Spectrum surfColor = sc.InitializeColor(RI_CS, Spectrum(1.));
  //tmap = sc.GetTexture(texWings[0]->texture);
  s = sc.InitializeFloat(RI_S, 0.);
  t = sc.InitializeFloat(RI_T, 0.);

 if( identity[1] == 'r' )
   surfColor = wingTex->Lookup(1-s, t, false );
   else
     surfColor = wingTex->Lookup(s, t, false );
  
  // HACK!!
  //if( strcmp( identity, "ll" ) == 0 )
  //  surfColor = Spectrum( .5, .5, .5 );

  return PlasticShade(sc, surfColor );

}


ShadeContext::ShadeContext(const HitInfo * hi, const Vector & w)
:wo(w.Hat())
{
	hitInfo = hi;

	Ng = hitInfo->hitPrim->attributes->ObjectToWorld(hitInfo->NgObj);
	Ng.Normalize();
	if (Dot(Ng, wo) < 0.) {
		Ng = -Ng;
		entering = false;
	} else {
		entering = true;
	}

	const Normal *Ni = hitInfo->GetNormal(RI_N);
	if (Ni != NULL) {
		Ns = hitInfo->hitPrim->attributes->ObjectToWorld(*Ni).Hat();
		if (Dot(Ns, wo) < 0.)
			Ns = -Ns;
	} else {
		Ns = Ng;
	}

}

Float ShadeContext::InitializeFloat(RtToken token, Float def) const
{
	const Float *ret = hitInfo->GetFloat(token);
	if (ret)
		return *ret;
	if (token == RI_S)
		return hitInfo->u;
	else if (token == RI_T)
		return hitInfo->v;
	return def;
}

Spectrum ShadeContext::InitializeColor(RtToken token, const Spectrum & def) const
{
	const Spectrum *ret = hitInfo->GetColor(token);
	if (ret)
		return *ret;
	if (token == RI_CS)
		return hitInfo->hitPrim->attributes->Color;
	else if (token == RI_OS)
		return hitInfo->hitPrim->attributes->Opacity;
	return def;
}

Point ShadeContext::InitializePoint(RtToken token, const Point & def) const
{
	const Point *ret = hitInfo->GetPoint(token);
	if (ret)
		return *ret;
	else
		return def;
}

const char *ShadeContext::InitializeString(RtToken token, const char *def) const
{
	const char *ret = hitInfo->GetString(token);
	if (ret)
		return ret;
	else
		return def;
}

Vector ShadeContext::InitializeVector(RtToken token, const Vector & def) const
{
	const Vector *ret = hitInfo->GetVector(token);
	if (ret)
		return *ret;
	else
		return def;
}

Normal ShadeContext::InitializeNormal(RtToken token, const Normal & def) const
{
	const Normal *ret = hitInfo->GetNormal(token);
	if (ret)
		return *ret;
	else
		return def;
}

StringHashTable ShadeContext::textures;


TextureMap *ShadeContext::GetTexture(const char *filename)
{
	TextureMap *ret = (TextureMap *) textures.Search(filename);
	if (!ret) {
		ret = new TextureMap(filename);
		textures.Add(filename, ret);
	}
	return ret;
}

Material::~Material()
{
	delete surfaceFunction;
}

BRDF *MatteSurface::Shade(const ShadeContext & sc) const
{
	Spectrum Cs = sc.InitializeColor(RI_CS, Spectrum(1.0));
	Float surfaceKd = sc.InitializeFloat(RI_KD, Kd);
	return CreateLambertian(surfaceKd * Cs, sc.Ns);
}

BRDF *PlasticSurface::Shade(const ShadeContext & sc) const
{
	Spectrum Cs = sc.InitializeColor(RI_CS, Spectrum(1.0));
	return PlasticShade(sc, Cs);
}

BRDF *PlasticSurface::PlasticShade(const ShadeContext & sc,
								   const Spectrum & Cs) const
{

	Float surfKd = sc.InitializeFloat(RI_KD, Kd);
	Float surfKs = sc.InitializeFloat(RI_KS, Ks);
	Float surfRoughness = sc.InitializeFloat(RI_ROUGHNESS, Roughness);
	Spectrum surfSpecColor =
		sc.InitializeColor(RI_SPECULARCOLOR, SpecularColor);

	ScatteringMixture *mix = CreateScatteringMixture();
	mix->AddFunction(CreateLambertian(surfKd * Cs, sc.Ns));
	mix->AddFunction(CreateBlinnGlossy(surfKs * surfSpecColor,
					   Clamp(surfRoughness, 0., 1.), sc.Ns,
					   sc.wo));
	return mix;

}

PaintedPlasticSurface::PaintedPlasticSurface(SurfaceFunction *data)
		: PlasticSurface(data) {
	texName = surfaceFunction->InitializeString(RI_TEXTURENAME,
								NULL);
}

BRDF *PaintedPlasticSurface::Shade(const ShadeContext & sc) const
{
	Spectrum surfColor = sc.InitializeColor(RI_CS, Spectrum(1.));
	const char *surfTexName = sc.InitializeString(RI_TEXTURENAME, texName);
	TextureMap *tmap = sc.GetTexture(surfTexName);

	if (tmap) {
		Float s = sc.InitializeFloat(RI_S, 0.);
		Float t = sc.InitializeFloat(RI_T, 0.);
		surfColor *= tmap->Lookup(s, t, false);
	}
	return PlasticShade(sc, surfColor);
}


BRDF *CheckeredSurface::Shade(const ShadeContext & sc) const
{

	Float surfKd = sc.InitializeFloat(RI_KD, Kd);
	Spectrum surfCColor1 = sc.InitializeColor(LRT_CHECKCOLOR1,
											  CheckColor1);
	Spectrum surfCColor2 = sc.InitializeColor(LRT_CHECKCOLOR2,
											  CheckColor2);
	Float freq = sc.InitializeFloat(LRT_CHECKFREQUENCY,
									CheckFrequency);

	Point Pshade = sc.InitializePoint(RI_P, Point(0, 0, 0));
	Point Pcheck(freq * Pshade.x, freq * Pshade.y, freq * Pshade.z);
	if (Mod(Round(Pcheck.x) + Round(Pcheck.y) + Round(Pcheck.z), 2) == 0) {
		return CreateLambertian(surfKd * surfCColor1, sc.Ns);
	} else {
		return CreateLambertian(surfKd * surfCColor2, sc.Ns);
	}

}

BRDF *GlassSurface::Shade(const ShadeContext & sc) const
{

	Float shadeKr = sc.InitializeFloat(RI_KR, Kr);
	Float shadeKt = sc.InitializeFloat(LRT_KT, Kt);
	Float shadeIndex = sc.InitializeFloat(LRT_INDEX, Index);
	Spectrum Cs = sc.InitializeColor(RI_CS, Spectrum(0.));

	if (shadeKr == 0. && shadeKt == 0.)
		return NULL;

	ScatteringMixture *mix = CreateScatteringMixture();
	if (shadeKr > 0.) {
		mix->AddFunction(CreateSpecularReflection(shadeKr * Cs,
												  sc.Ns, sc.wo));
	}
	if (shadeKt > 0.) {

		if (sc.entering) {
			mix->AddFunction(CreateSpecularTransmission(Kt * Cs,
														sc.Ns, sc.wo,
														1.0, shadeIndex));
		} else {
			mix->AddFunction(CreateSpecularTransmission(Kt * Cs,
														sc.Ns, sc.wo,
														shadeIndex, 1.0));
		}

	}
	return mix;
}

BRDF *ShinyMetalSurface::Shade(const ShadeContext & sc) const
{

	Float shadeKs = sc.InitializeFloat(RI_KS, Ks);
	Float shadeKr = sc.InitializeFloat(RI_KR, Kr);
	Float shadeRoughness = sc.InitializeFloat(RI_ROUGHNESS, Roughness);
	Spectrum Cs = sc.InitializeColor(RI_CS, Spectrum(1.0));

	ScatteringMixture *mix = CreateScatteringMixture();
	if (shadeKs > 0.) {
		mix->AddFunction(CreateBlinnGlossy(shadeKs * Cs,
										   shadeRoughness, sc.Ns, sc.wo));
	}
	if (shadeKr > 0.) {
		mix->AddFunction(CreateSpecularReflection(shadeKr * Cs,
												  sc.Ns, sc.wo));
	}
	return mix;
}

BRDF *ViewSTSurface::Shade(const ShadeContext & sc) const
{
	Float sp = sc.InitializeFloat(RI_S, 0.);
	Float tp = sc.InitializeFloat(RI_T, 0.);
	return CreateLambertian(Spectrum(sp, tp, 0.), sc.Ns);
}

Material *MaterialCreate(const char *name, SurfaceFunction * bindings)
{
	if (strcmp(name, RI_MATTE) == 0)
		return new MatteSurface(bindings);
	else if (strcmp(name, RI_PLASTIC) == 0)
		return new PlasticSurface(bindings);
	else if (strcmp(name, RI_PAINTEDPLASTIC) == 0)
		return new PaintedPlasticSurface(bindings);
	else if (strcmp(name, RI_WINGSURFACE) == 0 )
	        return new WingSurface(NULL,bindings);
	else if (strcmp(name, RI_SHINYMETAL) == 0)
		return new ShinyMetalSurface(bindings);
	else if (strcmp(name, LRT_VIEWST) == 0)
		return new ViewSTSurface(bindings);
	else if (strcmp(name, LRT_CHECKERED) == 0)
		return new CheckeredSurface(bindings);
	else if (strcmp(name, LRT_GLASS) == 0)
		return new GlassSurface(bindings);
	else {
		Warning("Unknown surface shader: %s," "returning \"matte\"", name);
		return new MatteSurface(bindings);
	}
}

SurfaceFunction::~SurfaceFunction()
{

	for (u_int i = 0; i < vertexPoints.size(); ++i)
		delete[]vertexPoints[i].second;

	for (u_int i = 0; i < uniformStrings.size(); ++i)
		delete uniformStrings[i].second;

	for (u_int i = 0; i < vertexFloats.size(); ++i)
		delete[]vertexFloats[i].second;
	for (u_int i = 0; i < vertexHPoints.size(); ++i)
		delete[]vertexHPoints[i].second;
	for (u_int i = 0; i < vertexVectors.size(); ++i)
		delete[]vertexVectors[i].second;
	for (u_int i = 0; i < vertexNormals.size(); ++i)
		delete[]vertexNormals[i].second;
	for (u_int i = 0; i < vertexColors.size(); ++i)
		delete[]vertexColors[i].second;

}

SurfaceFunction::SurfaceFunction(const SurfaceFunction & sf)
{
	nVertex = sf.nVertex;

	vector<pair<RtToken, Point> >::const_iterator uniformPointIter;
	for (uniformPointIter = sf.uniformPoints.begin();
		 uniformPointIter != sf.uniformPoints.end(); uniformPointIter++) {
		uniformPoints.push_back(make_pair((*uniformPointIter).first,
										  (*uniformPointIter).second));
	}

	vector<pair<RtToken, Point *> >::const_iterator vertexPointIter;
	for (vertexPointIter = sf.vertexPoints.begin();
		 vertexPointIter != sf.vertexPoints.end(); vertexPointIter++) {
		Point *values = new Point[nVertex];
		memcpy(values,
			   (*vertexPointIter).second, nVertex * sizeof(*values));
		vertexPoints.
			push_back(make_pair((*vertexPointIter).first, values));
	}

	vector<pair<RtToken, Float> >::const_iterator uniformFloatIter;
	for (uniformFloatIter = sf.uniformFloats.begin();
		 uniformFloatIter != sf.uniformFloats.end(); uniformFloatIter++) {
		uniformFloats.push_back(make_pair((*uniformFloatIter).first,
										  (*uniformFloatIter).second));
	}
	vector<pair<RtToken, Float> >::const_iterator uniformHPointIter;
	for (uniformHPointIter = sf.uniformHPoints.begin();
		 uniformHPointIter != sf.uniformHPoints.end(); uniformHPointIter++) {
		uniformHPoints.push_back(make_pair((*uniformHPointIter).first,
										   (*uniformHPointIter).second));
	}
	vector<pair<RtToken, Vector> >::const_iterator uniformVectorIter;
	for (uniformVectorIter = sf.uniformVectors.begin();
		 uniformVectorIter != sf.uniformVectors.end(); uniformVectorIter++) {
		uniformVectors.push_back(make_pair((*uniformVectorIter).first,
										   (*uniformVectorIter).second));
	}
	vector<pair<RtToken, Normal> >::const_iterator uniformNormalIter;
	for (uniformNormalIter = sf.uniformNormals.begin();
		 uniformNormalIter != sf.uniformNormals.end(); uniformNormalIter++) {
		uniformNormals.push_back(make_pair((*uniformNormalIter).first,
										   (*uniformNormalIter).second));
	}
	vector<pair<RtToken, Spectrum> >::const_iterator uniformColorIter;
	for (uniformColorIter = sf.uniformColors.begin();
		 uniformColorIter != sf.uniformColors.end(); uniformColorIter++) {
		uniformColors.push_back(make_pair((*uniformColorIter).first,
										  (*uniformColorIter).second));
	}
	vector<pair<RtToken, char *> >::const_iterator uniformStringIter;
	for (uniformStringIter = sf.uniformStrings.begin();
		 uniformStringIter != sf.uniformStrings.end(); uniformStringIter++) {
		uniformStrings.push_back(make_pair((*uniformStringIter).first,
										   Strdup((*uniformStringIter).
												  second)));
	}

	vector<pair<RtToken, Float *> >::const_iterator vertexFloatIter;
	for (vertexFloatIter = sf.vertexFloats.begin();
		 vertexFloatIter != sf.vertexFloats.end(); vertexFloatIter++) {
		Float *values = new Float[nVertex];
		memcpy(values, (*vertexFloatIter).second,
			   nVertex * sizeof(*values));
		vertexFloats.
			push_back(make_pair((*vertexFloatIter).first, values));
	}
	vector<pair<RtToken, Float *> >::const_iterator vertexHPointIter;
	for (vertexHPointIter = sf.vertexHPoints.begin();
		 vertexHPointIter != sf.vertexHPoints.end(); vertexHPointIter++) {
		Float *values = new Float[4 * nVertex];
		memcpy(values, (*vertexHPointIter).second,
			   4 * nVertex * sizeof(*values));
		vertexHPoints.
			push_back(make_pair((*vertexHPointIter).first, values));
	}
	vector<pair<RtToken, Vector *> >::const_iterator vertexVectorIter;
	for (vertexVectorIter = sf.vertexVectors.begin();
		 vertexVectorIter != sf.vertexVectors.end(); vertexVectorIter++) {
		Vector *values = new Vector[nVertex];
		memcpy(values, (*vertexVectorIter).second,
			   nVertex * sizeof(*values));
		vertexVectors.
			push_back(make_pair((*vertexVectorIter).first, values));
	}
	vector<pair<RtToken, Normal *> >::const_iterator vertexNormalIter;
	for (vertexNormalIter = sf.vertexNormals.begin();
		 vertexNormalIter != sf.vertexNormals.end(); vertexNormalIter++) {
		Normal *values = new Normal[nVertex];
		memcpy(values, (*vertexNormalIter).second,
			   nVertex * sizeof(*values));
		vertexNormals.
			push_back(make_pair((*vertexNormalIter).first, values));
	}
	vector<pair<RtToken, Spectrum *> >::const_iterator vertexColorIter;
	for (vertexColorIter = sf.vertexColors.begin();
		 vertexColorIter != sf.vertexColors.end(); vertexColorIter++) {
		Spectrum *values = new Spectrum[nVertex];
		memcpy(values, (*vertexColorIter).second,
			   nVertex * sizeof(*values));
		vertexColors.
			push_back(make_pair((*vertexColorIter).first, values));
	}

}

void SurfaceFunction::AddUniformPoint(RtToken name, const Float * data)
{
	uniformPoints.
		push_back(make_pair(name, Point(data[0], data[1], data[2])));
}

void SurfaceFunction::AddVertexPoint(RtToken name, const Float * data)
{
	Assert(nVertex > 0);
	Point *values = new Point[nVertex];
	for (u_int i = 0; i < nVertex; ++i) {
		values[i].x = data[3 * i];
		values[i].y = data[3 * i + 1];
		values[i].z = data[3 * i + 2];
	}
	vertexPoints.push_back(make_pair(name, values));
}

void SurfaceFunction::AddUniformFloat(RtToken name, const Float * data)
{
	uniformFloats.push_back(make_pair(name, *data));
}
void SurfaceFunction::AddUniformHPoint(RtToken name, const Float * data)
{
	uniformHPoints.push_back(make_pair(name, data[0]));
	uniformHPoints.push_back(make_pair(name, data[1]));
	uniformHPoints.push_back(make_pair(name, data[2]));
	uniformHPoints.push_back(make_pair(name, data[3]));
}
void SurfaceFunction::AddUniformVector(RtToken name, const Float * data)
{
	uniformVectors.
		push_back(make_pair(name, Vector(data[0], data[1], data[2])));
}
void SurfaceFunction::AddUniformNormal(RtToken name, const Float * data)
{
	uniformNormals.
		push_back(make_pair(name, Normal(data[0], data[1], data[2])));
}
void SurfaceFunction::AddUniformColor(RtToken name, const Float * data)
{
	uniformColors.
		push_back(make_pair(name, Spectrum(data[0], data[1], data[2])));
}
void SurfaceFunction::AddUniformString(RtToken name, const char *data)
{
	uniformStrings.push_back(make_pair(name, Strdup(data)));
}

void SurfaceFunction::AddVertexFloat(RtToken name, const Float * data)
{
	Assert(nVertex > 0);
	Float *values = new Float[nVertex];
	memcpy(values, data, nVertex * sizeof(Float));
	vertexFloats.push_back(make_pair(name, values));
}
void SurfaceFunction::AddVertexHPoint(RtToken name, const Float * data)
{
	Assert(nVertex > 0);
	Float *values = new Float[4 * nVertex];
	memcpy(values, data, 4 * nVertex * sizeof(Float));
	vertexHPoints.push_back(make_pair(name, values));
}
void SurfaceFunction::AddVertexVector(RtToken name, const Float * data)
{
	Assert(nVertex > 0);
	Vector *values = new Vector[nVertex];
	for (u_int i = 0; i < nVertex; ++i) {
		values[i].x = data[3 * i];
		values[i].y = data[3 * i + 1];
		values[i].z = data[3 * i + 2];
	}
	vertexVectors.push_back(make_pair(name, values));
}
void SurfaceFunction::AddVertexNormal(RtToken name, const Float * data)
{
	Assert(nVertex > 0);
	Normal *values = new Normal[nVertex];
	for (u_int i = 0; i < nVertex; ++i) {
		values[i].x = data[3 * i];
		values[i].y = data[3 * i + 1];
		values[i].z = data[3 * i + 2];
	}
	vertexNormals.push_back(make_pair(name, values));
}
void SurfaceFunction::AddVertexColor(RtToken name, const Float * data)
{
	Assert(nVertex > 0);
	Spectrum *values = new Spectrum[nVertex];
	for (u_int i = 0; i < nVertex; ++i)
		values[i] =
			Spectrum(data[3 * i], data[3 * i + 1], data[3 * i + 2]);
	vertexColors.push_back(make_pair(name, values));
}

const Point *SurfaceFunction::GetUniformPoint(RtToken name) const
{
	for (u_int i = 0; i < uniformPoints.size(); ++i) {
		if (name == uniformPoints[i].first)
			return &uniformPoints[i].second;
	}
	return NULL;
}

const Point *SurfaceFunction::GetVertexPoint(RtToken name) const
{
	for (u_int i = 0; i < vertexPoints.size(); ++i) {
		if (name == vertexPoints[i].first)
			return vertexPoints[i].second;
	}
	return NULL;
}

const Float *SurfaceFunction::GetUniformFloat(RtToken name) const
{
	for (u_int i = 0; i < uniformFloats.size(); ++i) {
		if (name == uniformFloats[i].first)
			return &uniformFloats[i].second;
	}
	return NULL;
}
const Vector *SurfaceFunction::GetUniformVector(RtToken name) const
{
	for (u_int i = 0; i < uniformVectors.size(); ++i) {
		if (name == uniformVectors[i].first)
			return &uniformVectors[i].second;
	}
	return NULL;
}
const Normal *SurfaceFunction::GetUniformNormal(RtToken name) const
{
	for (u_int i = 0; i < uniformNormals.size(); ++i) {
		if (name == uniformNormals[i].first)
			return &uniformNormals[i].second;
	}
	return NULL;
}
const Spectrum *SurfaceFunction::GetUniformColor(RtToken name) const
{
	for (u_int i = 0; i < uniformColors.size(); ++i) {
		if (name == uniformColors[i].first)
			return &uniformColors[i].second;
	}
	return NULL;
}
const char *SurfaceFunction::GetUniformString(RtToken name) const
{
	for (u_int i = 0; i < uniformStrings.size(); ++i) {
		if (name == uniformStrings[i].first)
			return uniformStrings[i].second;
	}
	return NULL;
}

const Float *SurfaceFunction::GetVertexFloat(RtToken name) const
{
	for (u_int i = 0; i < vertexFloats.size(); ++i) {
		if (name == vertexFloats[i].first)
			return vertexFloats[i].second;
	}
	return NULL;
}
const Float *SurfaceFunction::GetVertexHPoint(RtToken name) const
{
	for (u_int i = 0; i < vertexHPoints.size(); ++i) {
		if (name == vertexHPoints[i].first)
			return vertexHPoints[i].second;
	}
	return NULL;
}
const Vector *SurfaceFunction::GetVertexVector(RtToken name) const
{
	for (u_int i = 0; i < vertexVectors.size(); ++i) {
		if (name == vertexVectors[i].first)
			return vertexVectors[i].second;
	}
	return NULL;
}
const Normal *SurfaceFunction::GetVertexNormal(RtToken name) const
{
	for (u_int i = 0; i < vertexNormals.size(); ++i) {
		if (name == vertexNormals[i].first)
			return vertexNormals[i].second;
	}
	return NULL;
}
const Spectrum *SurfaceFunction::GetVertexColor(RtToken name) const
{
	for (u_int i = 0; i < vertexColors.size(); ++i) {
		if (name == vertexColors[i].first)
			return vertexColors[i].second;
	}
	return NULL;
}

Point SurfaceFunction::InitializePoint(RtToken name, const Point & def) const
{
	const Point *ptr = GetUniformPoint(name);
	if (ptr != NULL)
		return *ptr;
	else
		return def;
}

Float SurfaceFunction::InitializeFloat(RtToken name, Float def) const
{
	const Float *ptr = GetUniformFloat(name);
	if (ptr != NULL)
		return *ptr;
	else
		return def;
}

Spectrum SurfaceFunction::InitializeColor(RtToken name,
										  const Spectrum & def) const
{
	const Spectrum *ptr = GetUniformColor(name);
	if (ptr != NULL)
		return *ptr;
	else
		return def;
}

Vector SurfaceFunction::InitializeVector(RtToken name, const Vector & def) const
{
	const Vector *ptr = GetUniformVector(name);
	if (ptr != NULL)
		return *ptr;
	else
		return def;
}

Normal SurfaceFunction::InitializeNormal(RtToken name, const Normal & def) const
{
	const Normal *ptr = GetUniformNormal(name);
	if (ptr != NULL)
		return *ptr;
	else
		return def;
}
const char *SurfaceFunction::InitializeString(RtToken name,
											  const char *def) const
{
	const char *ptr = GetUniformString(name);
	if (ptr != NULL)
		return ptr;
	else
		return def;
}

InterpolatedPrimData::InterpolatedPrimData(const SurfaceFunction * pd)
{
	primitiveData = pd;

	if (primitiveData->vertexPoints.size() > 0) {
		int nvp = primitiveData->vertexPoints.size();
		interpVertexPoints = new Point[nvp];
	} else
		interpVertexPoints = NULL;

	if (primitiveData->vertexFloats.size() > 0) {
		interpVertexFloats = new Float[primitiveData->vertexFloats.size()];
	} else {
		interpVertexFloats = NULL;
	}
	if (primitiveData->vertexVectors.size() > 0) {
		int nv = primitiveData->vertexVectors.size();
		interpVertexVectors = new Vector[nv];
	} else {
		interpVertexVectors = NULL;
	}
	if (primitiveData->vertexNormals.size() > 0) {
		int vn = primitiveData->vertexNormals.size();
		interpVertexNormals = new Normal[vn];
	} else {
		interpVertexNormals = NULL;
	}
	if (primitiveData->vertexColors.size() > 0) {
		int nc = primitiveData->vertexColors.size();
		interpVertexColors = new Spectrum[nc];
	} else {
		interpVertexColors = NULL;
	}

}

InterpolatedPrimData::~InterpolatedPrimData()
{

	delete[]interpVertexPoints;

	delete[]interpVertexFloats;
	delete[]interpVertexVectors;
	delete[]interpVertexNormals;
	delete[]interpVertexColors;

}

InterpolatedPrimData *SurfaceFunction::Interpolate(int n,
												   const int *offsets,
												   const Float * weights) const
{
	InterpolatedPrimData *ret = new InterpolatedPrimData(this);

	for (u_int i = 0; i < vertexPoints.size(); ++i) {
		Point value(0, 0, 0);
		Point *pts = vertexPoints[i].second;
		for (int j = 0; j < n; ++j) {
			int o = offsets[j];
			value.x += weights[j] * pts[o].x;
			value.y += weights[j] * pts[o].y;
			value.z += weights[j] * pts[o].z;
		}
		ret->interpVertexPoints[i] = value;
	}

	for (u_int i = 0; i < vertexFloats.size(); ++i) {
		Float value = 0.;
		Float *f = vertexFloats[i].second;
		for (int j = 0; j < n; ++j) {
			int o = offsets[j];
			value += weights[j] * f[o];
		}
		ret->interpVertexFloats[i] = value;
	}
	for (u_int i = 0; i < vertexVectors.size(); ++i) {
		Vector value(0, 0, 0);
		Vector *vec = vertexVectors[i].second;
		for (int j = 0; j < n; ++j) {
			int o = offsets[j];
			value += weights[j] * vec[o];
		}
		ret->interpVertexVectors[i] = value;
	}
	for (u_int i = 0; i < vertexNormals.size(); ++i) {
		Normal value(0, 0, 0);
		Normal *norm = vertexNormals[i].second;
		for (int j = 0; j < n; ++j) {
			int o = offsets[j];
			value += weights[j] * norm[o];
		}
		ret->interpVertexNormals[i] = value;
	}
	for (u_int i = 0; i < vertexColors.size(); ++i) {
		Spectrum value(0.);
		Spectrum *cp = vertexColors[i].second;
		for (int j = 0; j < n; ++j) {
			int o = offsets[j];
			value += weights[j] * cp[o];
		}
		ret->interpVertexColors[i] = value;
	}

	return ret;
}

const Point *InterpolatedPrimData::GetPoint(RtToken token) const
{
	const Point *ret = primitiveData->GetUniformPoint(token);
	if (ret)
		return ret;

	const vector<pair<RtToken, Point *> > &vpp =
	primitiveData->vertexPoints;
	for (u_int i = 0; i < vpp.size(); ++i) {
		if (token == vpp[i].first)
			return &interpVertexPoints[i];
	}
	return NULL;

}

const Float *InterpolatedPrimData::GetFloat(RtToken token) const
{
	const vector<pair<RtToken, Float *> > &vfp = primitiveData->vertexFloats;
	for (u_int i = 0; i < vfp.size(); ++i) {
		if (token == vfp[i].first)
			return &interpVertexFloats[i];
	}
	return NULL;
}
const Vector *InterpolatedPrimData::GetVector(RtToken token) const
{
	const vector<pair<RtToken, Vector *> > &vvp = primitiveData->vertexVectors;
	for (u_int i = 0; i < vvp.size(); ++i) {
		if (token == vvp[i].first)
			return &interpVertexVectors[i];
	}
	return NULL;
}
const Normal *InterpolatedPrimData::GetNormal(RtToken token) const
{
	const vector<pair<RtToken, Normal *> > &vnp = primitiveData->vertexNormals;
	for (u_int i = 0; i < vnp.size(); ++i) {
		if (token == vnp[i].first)
			return &interpVertexNormals[i];
	}
	return NULL;
}
const Spectrum *InterpolatedPrimData::GetColor(RtToken token) const
{
	const vector<pair<RtToken, Spectrum *> > &vcp = primitiveData->vertexColors;
	for (u_int i = 0; i < vcp.size(); ++i) {
		if (token == vcp[i].first)
			return &interpVertexColors[i];
	}
	return NULL;
}

float WingSurface::RandomRange( float min, float max ){
  
 
  //float test = (float)rand() / (float) RAND_MAX;
  //float test2 = max - min;
  //float test3 = test * test2;
  //float test4 = test3 + min;
return ((float)rand() / (float)RAND_MAX) * (max - min) + min;
}

WingSurface::WingSurface( Butterfly *buttergeo, SurfaceFunction *data)
		: PlasticSurface(data) {
  
  bgeo = buttergeo;
  RGBATex *spot, *mask;
 

  /*texWings.push_back( new RGBATex( "planewing.tif",
    "planewing_alpha.tif") );*/

  spot = new RGBATex("spots/bluespot.tif", "spots/bluespot_alpha.tif");
  spot->trans = Identity();
  texSpots.push_back( spot );
    
  spot = new RGBATex("spots/spot2.tif", "spots/spot2_alpha.tif");
  spot->trans = Identity();
  texSpots.push_back( spot );

  spot = new RGBATex("spots/bluespot.tif", "spots/bluespot_alpha.tif");
  spot->trans = Identity();
  texSpots.push_back( spot );

  spot = new RGBATex("spots/spot3.tif", "spots/spot3_alpha.tif");
  spot->trans = Identity();
  texSpots.push_back( spot );

  spot = new RGBATex("spots/spot4.tif", "spots/spot4_alpha.tif");
  spot->trans = Identity();
  texSpots.push_back( spot );



  // MASKS 

  mask = new RGBATex("black.tif", "spots/mask1.tif");
  mask->trans = Identity();
  texMasks.push_back( mask );

  
  mask = new RGBATex("noalpha.tif", "spots/mask2.tif");
  mask->trans = Identity();
  texMasks.push_back( mask );
  
  mask = new RGBATex("black.tif", "spots/mask3.tif");
  mask->trans = Identity();
  texMasks.push_back( mask );
  
  mask = new RGBATex("black.tif", "spots/mask4.tif");
  mask->trans = Identity();
  texMasks.push_back( mask );
  
  
  mask = new RGBATex("black.tif", "spots/mask5.tif");
  mask->trans = Identity();
  texMasks.push_back( mask );
  
  mask = new RGBATex("black.tif", "spots/mask6.tif");
  mask->trans = Identity();
  texMasks.push_back( mask );

  mask = new RGBATex("black.tif", "spots/mask7.tif");
  mask->trans = Identity();
  texMasks.push_back( mask );
   
}






